"""Imports for Python API.

This file is MACHINE GENERATED! Do not edit.
Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""
from tensorflow.python.saved_model.utils import build_tensor_info
from tensorflow.python.saved_model.utils import get_tensor_from_tensor_info