"""Imports for Python API.

This file is MACHINE GENERATED! Do not edit.
Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""
from tensorflow.python.feature_column.feature_column import bucketized_column
from tensorflow.python.feature_column.feature_column import categorical_column_with_hash_bucket
from tensorflow.python.feature_column.feature_column import categorical_column_with_identity
from tensorflow.python.feature_column.feature_column import categorical_column_with_vocabulary_file
from tensorflow.python.feature_column.feature_column import categorical_column_with_vocabulary_list
from tensorflow.python.feature_column.feature_column import crossed_column
from tensorflow.python.feature_column.feature_column import embedding_column
from tensorflow.python.feature_column.feature_column import indicator_column
from tensorflow.python.feature_column.feature_column import input_layer
from tensorflow.python.feature_column.feature_column import linear_model
from tensorflow.python.feature_column.feature_column import make_parse_example_spec
from tensorflow.python.feature_column.feature_column import numeric_column
from tensorflow.python.feature_column.feature_column import shared_embedding_columns
from tensorflow.python.feature_column.feature_column import weighted_categorical_column