"""Imports for Python API.

This file is MACHINE GENERATED! Do not edit.
Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""
from tensorflow.python.keras._impl.keras.losses import KLD as kullback_leibler_divergence
from tensorflow.python.keras._impl.keras.losses import MAE as mean_absolute_error
from tensorflow.python.keras._impl.keras.losses import MAPE as mean_absolute_percentage_error
from tensorflow.python.keras._impl.keras.losses import MSE as mean_squared_error
from tensorflow.python.keras._impl.keras.losses import MSLE as mean_squared_logarithmic_error
from tensorflow.python.keras._impl.keras.losses import binary_crossentropy
from tensorflow.python.keras._impl.keras.losses import categorical_crossentropy
from tensorflow.python.keras._impl.keras.losses import cosine as cosine_proximity
from tensorflow.python.keras._impl.keras.losses import hinge
from tensorflow.python.keras._impl.keras.losses import poisson
from tensorflow.python.keras._impl.keras.losses import sparse_categorical_crossentropy
from tensorflow.python.keras._impl.keras.losses import squared_hinge
from tensorflow.python.keras._impl.keras.metrics import binary_accuracy
from tensorflow.python.keras._impl.keras.metrics import categorical_accuracy
from tensorflow.python.keras._impl.keras.metrics import deserialize
from tensorflow.python.keras._impl.keras.metrics import get
from tensorflow.python.keras._impl.keras.metrics import serialize
from tensorflow.python.keras._impl.keras.metrics import sparse_top_k_categorical_accuracy
from tensorflow.python.keras._impl.keras.metrics import top_k_categorical_accuracy