"""Imports for Python API.

This file is MACHINE GENERATED! Do not edit.
Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""
from tensorflow.python.keras._impl.keras.regularizers import L1L2
from tensorflow.python.keras._impl.keras.regularizers import Regularizer
from tensorflow.python.keras._impl.keras.regularizers import deserialize
from tensorflow.python.keras._impl.keras.regularizers import get
from tensorflow.python.keras._impl.keras.regularizers import l1
from tensorflow.python.keras._impl.keras.regularizers import l1_l2
from tensorflow.python.keras._impl.keras.regularizers import l2
from tensorflow.python.keras._impl.keras.regularizers import serialize